# Copyright (C) 2010-2019 by Dr. Dieter Maurer <dieter@handshake.de>
"""schema related utilities."""
from zope.interface import Interface, providedBy, implementer
from zope.interface.interfaces import IMethod
from zope.schema.interfaces import IField, IObject, \
     WrongContainedType, SchemaNotProvided, ValidationError, \
     SchemaNotFullyImplemented, \
     ITextLine
from zope.schema import Object as ObjectBase, Bool, getFieldsInOrder, \
     TextLine

from .interfaces import ISchemaConfigured

from dm.reuse.proxy import make_proxy


def interface_from_spec(spec, name=None, **kw):
  """build an interface from interface spec *spec*.

  This should probably go into its own package (maybe ``dm.zope.interface``).

  Note: the interfaces constructed in this way may not be picklable
  (not tested). If they are indeed not picklable, they should not be stored.
  """
  # try to guess whether "spec" is an interface or specification
  #  The implementors have broken "issubclass", we therefore need
  #  more indirect (and in principle less reliable) methods
  # if issubclass(spec, Interface):
  if hasattr(spec, 'names'):
    if name is None and not kw: return spec
    spec = (spec,)
  name = name is None and 'FromSpec' or name
  return type(Interface)(name, tuple(spec), kw)


def schemaitems(spec):
  """The schema part of interface specification *spec* as a list of id, field pairs."""
  iface = interface_from_spec(spec)
  # may want to filter duplicates out or raise an exception on duplicates
  seen = set(); items = []
  for (name, field) in getFieldsInOrder(iface):
    if name in seen: continue
    seen.add(name); items.append((name, field))
  return items


def schemadict(spec):
  """The schema part of interface specification *spec* as a ``dict``."""
  return dict(schemaitems(spec))
  

@implementer(ISchemaConfigured)
class SchemaConfigured(object):
  """Mixin class to provide configuration by the provided schema components."""

  def __init__(self, **kw):
    schema = schemadict(self.sc_schema_spec())
    for k in kw:
      # might want to control this check
      if k not in schema:
        raise TypeError('non schema keyword argument: %s' % k)
      setattr(self, k, kw[k])
    # provide default values for schema fields not set
    for f in schema:
      if not hasattr(self, f): setattr(self, f, schema[f].default)

  # provide control over which interfaces define the data schema
  SC_SCHEMAS = None

  def sc_schema_spec(self):
    """the schema specification which determines the data schema.

    This is determined by `SC_SCHEMAS` and defaults to `providedBy(self)`.
    """
    spec = self.SC_SCHEMAS
    if spec is None: return providedBy(self)
    return spec


class SchemaConfiguredEvolution(object):
  """Mixin class to support schema evolution for `SchemaConfigured` subclasses.

  Must come early in the resulting class's `mro` to be effective.
  """
  def __setstate__(self, state):
    sst = getattr(super(SchemaConfigured, self), "__setstate__", None)
    if sst is None and isinstance(state, dict):
      self.__dict__.update(state) # default
    else: sst(state)
    SchemaConfigured.__init__(self) # provide default values for missing fields


class _IObject(IObject):
  """auxiliary interface for ``Object`` extension.

  Note: no support yet for internationalization.
  """
  check_declaration = Bool(
    title=u"check declaration",
    description=u"check that the validated object declares the provided interface",
    default=True,
    )


@implementer(_IObject)
class Object(ObjectBase):
  """improved ``zope.schema.Object``.

  Error reports for ``zope.schema.Object`` forget to specify the
  problematic field. This class adds them.

  Occasionally, we do not wish to check that the existance of an
  interface declaration. Make this controllable with a
  ``check_declaration`` property.
  """

  # work around stupid ``Field`` constructor
  def __init__(self, *args, **kw):
    if 'check_declaration' in kw:
      self.check_declaration = kw['check_declaration']
      del kw['check_declaration']
    super(Object, self).__init__(*args, **kw)

  def _validate(self, value):
    base_validate = super(Object, self)._validate
    validate = base_validate if self.check_declaration \
               else (
      # suppress the declaration check via a proxy
      #   the suppression works be forcing the check to unconditionally succeed
      make_proxy(self,
                 schema=make_proxy(self.schema,
                                   providedBy=lambda unused: True
                                   )
                 )._rebind_proxied_method(base_validate)
      )
    validate(value)
    

##########################################################################
## Additional fields -- added in version 3.1

class IFilesystemPath(ITextLine):
  """An absolute or relative path in the filesystem.

  May need to be properly encoded for the actual file system access
  (which Python hopefully does automatically).
  """

@implementer(IFilesystemPath)
class FilesystemPath(TextLine):
  """Field holding a relative or absolute path in the filesystem."""


class IIriRef(ITextLine):
  """an Internanationalized Resource Identifier (= IRI) reference (see RFC 3987).

  also known as `anyURI` by XML-schema.
  """

from re import compile
iri_re = compile(u"(?:"
                 # `iunreservered` (without `ucschar`) + `gen-delims` + `sub-delims`
                 u"[-a-zA-Z0-9._~:/?#[\\]@!$&'()*+,;=]|" # 
                 # `pct-encoded`
                 u"(?:%[0-9a-zA-Z]{2})|"
                 # a superset of `ucschar`
                 u"[^\x00-\x9f]"
                 u")*$")


@implementer(IIriRef)
class IriRef(TextLine):
  """Field holding an IRI reference (see RFC 3987)."""
  def _validate(self, value):
    """currently checks only that *value* does not contain invalid characters."""
    super(IriRef, self)._validate(value)
    if iri_re.match(value) is None:
      raise ValidationError(
        # may want internationalization below
        self.__name__, 'the IriRef contains invalid characters'
        ).with_field_and_value(self, value)



# Copyright (C) 2019 by Dr. Dieter Maurer <dieter@handshake.de>
"""Extensions for Zope2/Zope 4+."""

from zope.interface import implementer

from zope.schema.interfaces import INativeStringLine, ValidationError
from zope.schema import NativeStringLine
from OFS.ObjectManager import bad_id


class IItemName(INativeStringLine):
  """A name (also known as id) for an `OFS.interfaces.IItem`. """

@implementer(IItemName)
class ItemName(NativeStringLine):
  def _validate(self, value):
    super(ItemName, self)._validate(self, value)
    if bad_id(value):
      raise ValidationError(
        # may want internationalization
        self.__name__, "Bad character[s] in ItemName"
        ).with_field_and_value(self, value)


class IItemPath(INativeStringLine):
  """An item path represented as a `"/"` separated sequence of `IItemName`s."""

@implementer(IItemPath)
class ItemPath(NativeStringLine):
  def _validate(self, value):
    super(ItemPath, self)._validate(value)
    if bad_id(value.replace("/", "")):
      raise ValidationError(
        # may want internationalization
        self.__name__, "Bad character[s] in ItemPath"
        ).with_field_and_value(self, value)
      





# Copyright (C) 2011-2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""Elementary form template."""

from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

from . import zope_major_version

form_template = ViewPageTemplateFile(
  "form_template%s.pt" % ("" if zope_major_version <= 2 else "-4")
  )



# Copyright (C) 2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""Resources to test the package.

You need a Zope [2 or 4] setup and must ensure that this directories
`configure.zcml` is activated during startup.
"""
from zope.interface import Interface, implementer
from zope.schema import TextLine, Int

from OFS.SimpleItem import SimpleItem

from ...schema import SchemaConfigured
from ...propertymanager import PropertyManager

from ..constructor import \
     add_form_factory, SchemaConfiguredZmiAddForm
from ..form import SchemaConfiguredZmiDisplayForm, SchemaConfiguredZmiEditForm

class ISchema(Interface):
  title = TextLine(title=u"Title", default=u"")
  number = Int(title=u"Number", required=False)
  

@implementer(ISchema)
class TestSchemaConfigured(SimpleItem, SchemaConfigured):
  """An object to test `dm.zope.schema.z2.constructor`."""

  meta_type = "Test schema configured constructor"

  manage_options = (
    dict(label="View", action="manage_view"),
    dict(label="Edit", action="manage_edit"),
    ) + SimpleItem.manage_options

  SC_SCHEMAS = ISchema

  # for test purposes only -- usually, you would use `ClassSecurityInfo`
  __roles__ = None
  manage_view__roles__ = None # public
  manage_edit__roles__ = ["Manager"]

  def manage_view(self, REQUEST):
    """show the object."""
    return SchemaConfiguredZmiDisplayForm(self, REQUEST)()

  def manage_edit(self, REQUEST):
    """esit the object."""
    return SchemaConfiguredZmiEditForm(self, REQUEST)()


@implementer(ISchema)
class TestPropertyManager(SimpleItem, PropertyManager, SchemaConfigured):
  """An object to test `dm.zope.schema.propertymanager`."""

  meta_type = "Test schema configured property manager"

  manage_options = PropertyManager.manage_options + SimpleItem.manage_options

  SC_SCHEMAS = ISchema


def initialize(context):
  context.registerClass(
    TestSchemaConfigured,
    constructors=(add_form_factory(TestSchemaConfigured,
                                   form_class=SchemaConfiguredZmiAddForm
                                   ),
                  )
    )
  # register `TestPropertyManager` if `dm.zope.generate` is present
  try:
    from dm.zope.generate.constructor import \
         add_form_factory as pm_add_form_factory, \
         add_action_factory
  except ImportError: pass
  else:
    context.registerClass(
      TestPropertyManager,
      constructors=(pm_add_form_factory(TestPropertyManager),
                    add_action_factory(TestPropertyManager),
                    ),
      )




# Copyright (C) 2011-2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""Zope 2 support"""
from sys import version_info

# Python version compatibility
python_major_version = version_info.major

#  Zope version compatibility
try:
  from App.version_txt import getZopeVersion
  zope_major_version = getZopeVersion()[0]
except ImportError: zope_major_version = 4 # work aroung Zope [4] beta bug



# Copyright (C) 2011-2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""Zope 2 (product class) constructor support for ``SchemaConfigured``.

Zope 2 is extended by so called "products" (which really are only
specially registered extensions). Products can register classes
which thereby become available via the so  called "add list" of
the Zope management interface. The interface between the "add list"
and the construction of class instances is implemented by
so called "constructors", provided as parameters during the
class registration via ``registerClass``.

This module allows to generate such a constructor for
classes derived from ``dm.zope.schema.SchemaConfigured``.
It makes use of a ``zope.formlib`` form generated according to
the class' schema.
"""
from Acquisition import aq_base, aq_parent
from zope.schema import ASCIILine, TextLine
from zope.formlib.form import Fields
from zope.i18nmessageid import MessageFactory

try: from Products.Five.formlib.formbase import AddForm
except ImportError:
  # some Zope version dropped formlib" support from "Five"; try "five.formlib"
  from five.formlib.formbase import AddForm

from ..form import SchemaConfiguredFormMixin

from .form import ZmiMixin

_ = MessageFactory('dm_zope_schema')


class SchemaConfiguredAddForm(SchemaConfiguredFormMixin, AddForm):
  """Default add form class."""

  __id_is_field = False

  def __init__(self, context, request):
    """**ATTENTION** *context* is expected to be the object added
    not the container; the container is expected to be found in
    the acquisition context of *context*.
    """
    super(SchemaConfiguredAddForm, self).__init__(context, request)
    parent = aq_parent(context)
##    if hasattr(context, '__parent__'): parent = context.__parent__
##    elif hasattr(context, 'aq_parent'): parent = context.aq_parent
##    else: parent = None # should not happen
##    self.__parent__ = parent
    # set ``render_context`` for all fields different from ``id``
    # check whether there is an ``id`` field
    for f in self.form_fields:
      if f.__name__ == 'id':
        self.__id_is_field = True
        f.get_rendered = lambda form: context.getId()
      else: f.render_context = True
    if not self.__id_is_field:
      self.form_fields = Fields(IdField(__name__='id', title=u"id")) \
                         + self.form_fields

  def create(self, data):
    ob = self.context
    ob._setId(data["id"]); del data["id"]
    ob.__init__(**data)
    return aq_base(ob)

  def add(self, ob):
    parent = self.__parent__; id = ob.getId()
    parent._setObject(id, ob)
    self._finished_add = True
    ob = parent._getOb(id) # update acquisition context
    # reset self.context -- work around acquisition bug
    aq_base(self).context = ob
    return ob

  def nextURL(self): 
    return self.context.absolute_url() + '/manage_workspace'


class SchemaConfiguredZmiAddForm(ZmiMixin, SchemaConfiguredAddForm):
  zmi_support = False # makes no sense for the add form
  form_class = "zmi-add"


def add_form_factory(class_, label=None, description=None, form_class=SchemaConfiguredAddForm, template=None):
  """return an add form for *class_* based on *form_class*.

  The add form can be used as an (usualy, the) *constructors* element
  of the ``registerClass`` call.

  *title* and *description* are used to set the forms *title* and
  *description*. They should be unicode
  or ``zope.i18nmessageid.MessageFactory`` instances.
  *title* defaults to "Create instance of *class_*\ " (translated
  in the translation domain ``dm_zope_schema``) and *class_*\ .__doc__.

  If you need your own template (e.g. when you are in an environment where
  the standard template does not work), you can provide it via *template*.
  Probably, it should be a
  `Products.Five.pagetemplatefile.ViewPageTemplateFile`.
  `template.form_template` might be a candidate.` 
  """
  def add_form(self, REQUEST):
    " "
    ob = class_()
    # provide parent information
    if hasattr(ob, '__of__'): ob = ob.__of__(self)
    else: ob.__parent__ = self # should we use an adapter?
    # Note: *template* might be a descriptor which has effect only when
    #  it is assigned to a class
    if template is not None: 
      class fc(form_class): pass
      fc.template = template
    else: fc = form_class
    form = fc(ob, REQUEST)
    if hasattr(form, "__of__"): form = form.__of__(self)
    form.__parent__ = self
    if label is not None: form.label = label
    elif not form.label: form.label = _(
      u'Create instance',
      u'Create ${class_} instance',
      dict(class_=class_.__name__), # might need to unicodify the name
      )
    if description is not None: form.description = description
    elif not getattr(form, "description", "") and class_.__doc__:
      form.description = class_.__doc__
    # make the form public -- protection is garanteed by the product dispatcher
    form.__roles__ = None
    return form()
  add_form.__name__ = "add_" + class_.__name__
  return add_form


############################################################################
#  Zope version compatibility
from . import zope_major_version, python_major_version
IdField = TextLine if python_major_version >= 3 else ASCIILine



# Copyright (C) 2010-2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""Schema based properties definition.

Interface between Zope[2]'s ``OFS.PropertyManager.PropertyManager`` and
``zope.schema`` schemas.
"""

from sys import version_info

import OFS.PropertyManager as pmm
from ComputedAttribute import ComputedAttribute
from ExtensionClass import Base
from zope.interface import Interface, implementedBy, providedBy
from zope.schema import BytesLine, TextLine, Tuple, Text, Bytes, Bool, Int, Float

from .schema import schemadict, schemaitems

TitleField = BytesLine if version_info[0] <= 2 else TextLine


class _Descriptor(object):
  """auxiliary descriptor class for ``_properties`` access."""
  def __get__(self, instance, cls=None):
    if instance is None:
      # class level access
      assert cls.SC_SCHEMAS is not None, "you must define `SC_SCHEMAS`"
      return _properties_from_schema(schemaitems(cls.SC_SCHEMAS), None)
    else: return _properties_from_schema(schemaitems(instance.SC_SCHEMAS), instance)

  def __set__(self, instance, value):
    raise NotImplementedError('``_properties`` change not supported.')


class TitleSchema(Interface):
  """define the ``title`` property (required for ``PropertyManager`` subclasses).

  Note: for Python 2, this is defined as ``BytesLine``.
  Therefore, non ASCII values will likely not work reliable.
  """

  title = TitleField(title=u'title', default='')



class _PropertyManagerMetaclass(type(pmm.PropertyManager)):
  def __getattr__(pm, key):
    """look up default property values."""
    sd = schemadict(pm.SC_SCHEMAS)
    # should we raise ``AttributeError``, when ``default`` is ``None``?
    if key in sd: return sd[key].default
    raise AttributeError(key)


## This does not work in Zope2
##class PropertyManager(six.with_metaclass(_PropertyManagerMetaclass,
##                                     pmm.PropertyManager)
##                      ):
class PropertyManager(pmm.PropertyManager):
  """a non extensible ``PropertyManager`` with schema defined property types.

  For the moment, we lack selection support. We would need automatic
  "select variable" definition (in our own ``__getattr__``) to change
  this.
  """
  __metaclass__ = _PropertyManagerMetaclass

  # support for class level default property values
  #__metaclass__ = _PropertyManagerMetaclass

  _properties = _Descriptor()
    
  # override ``PropertyManager`` methods
  #  we are not extensible -- maybe, we should provide our own
  #  template instead, to get rid of the ``management_page_charset`` hackles???
  manage_propertiesForm=pmm.DTMLFile('dtml/properties', pmm.__dict__,
                                     property_extensible_schema__=False
                                     )

  def _not_extensible(*args):
    raise NotImplementedError('property schema not extensible')

  _setProperty = _not_extensible
  _delProperty = _not_extensible
  manage_addProperty = _not_extensible

if not isinstance(PropertyManager, _PropertyManagerMetaclass):
  # Python 3: it does not honour the ``__metadata__` hint
  PropertyManager = _PropertyManagerMetaclass(
    PropertyManager.__name__,
    (PropertyManager,),
    dict(PropertyManager.__dict__)
    )


def _properties_from_schema(sis, context):
  """derive property descriptions from schema items *sis*.

  *context* is used to access ``management_page_charset`` in order
  to determine whether we can use unicode properties.
  """
  mpc = getattr(context, 'management_page_charset', None)
  if callable(mpc): mpc = mpc()
  uprops = mpc is None or mpc.upper() == 'UTF-8'

  def u(t): return uprops and 'u' + t or t

  from zope.schema import \
       Text, TextLine, Bytes, BytesLine, Bool, Int, Float, List

  def pd_from_field(id, field):
    bt = isinstance(field, Tuple) and field.value_type or field
    if isinstance(bt, TextLine): ty = u('string')
    elif isinstance(bt, Text): ty = u('text')
    elif isinstance(bt, BytesLine): ty = 'string'
    elif isinstance(bt, Bytes): ty = 'text'
    elif isinstance(bt, Bool): ty = 'bool'
    elif isinstance(bt, Int): ty = 'int'
    elif isinstance(bt, Float): ty = 'float'
    else: raise NotImplementedError('property conversion not implemented for field: %s' % id)
    if isinstance(field, Tuple):
      if not ty.endswith('string'):
        raise NotImplementedError('property conversion not implemented for field: %s' % id)
      ty = ty[-6:] + lines
    # may want to encode unicode values using ``mpc``
    pd = dict(id=id, label=field.title or id,
              mode=field.readonly and 'r' or 'rw',
              type=ty,
              description=field.description,
              )
    return pd

  return tuple(pd_from_field(*i) for i in sis)



# Copyright (C) 2014-2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""Field tagging.

Often you would like to provide additional information for your fields -
usually with effects to presentation. Unfortunately, `zope.schema` has
forgotten this use case: it is really hard to give existing fields
additional properties.

Fortunately, `zope.schema` is build on top of `zope.interface` and it
supports so called "tagged values" to add additional information to
its elements. This module uses this feature to provide additional information
for schema elements. As a matter of fact, it works for all schema elements.

Note: If your schemas are used across different packages, ensure the
tag names are sufficiently specific to avoid name clashes, e.g.
by using an appropriate prefix.
"""
from decorator import decorator


class Tagger(object):
  """Auxiliary class to tag `zope.interface` elements based on a prefix."""
  @decorator
  def wrap(f, self, elem, *args, **kw):
    return getattr(self.wrap(elem), f.__name__)(*args, **kw)

  def __init__(self, prefix=''):
    """*prefix* must not contain `:`"""
    self.__prefix = prefix

  @wrap
  def __call__(self, elem, **kw):
    """add *kw* tags to *elem*; return *elem*."""

  @wrap
  def get(self, elem, tag, default=None):
    """return value for *tag* or *default*."""

  @wrap
  def set(self, elem, tag, value):
    """set value of *tag* to *value*."""

  @wrap
  def list(self, elem):
    """list tags defined for *elem*."""

  @wrap
  def contains(self, elem, tag):
    """does *elem* has *tag*."""

  # we override `wrap`
  def wrap(self, elem): return TaggedElement(elem, self.__prefix)


class TaggedElement(object):
  """A `zope.interface.Element` wrapped for easy tagging/tag access."""
  def __init__(self, elem, prefix=''):
    """wrap *elem* for tagging/tag access with *prefix*.

    *prefix* must not contain `:`.
    """
    self.__elem = elem
    assert ":" not in prefix
    if prefix: prefix += ":"
    self.__prefix = prefix

  def __call__(self, **kw):
    """add tags described by *kw*; return the element."""
    elem, prefix = self.__elem, self.__prefix
    for k in kw: elem.setTaggedValue(prefix + k, kw[k])
    return elem

  def get(self, tag, default=None):
    elem, prefix = self.__elem, self.__prefix
    r = elem.queryTaggedValue(prefix + tag, default)
    if r is default and type(default) is type(KeyError) and issubclass(default, KeyError): raise default(tag)
    return r

  def __getitem__(self, tag): return self.get(tag, KeyError)

  def set(self, tag, value): self(tag=value)
  def contains(self, tag): return self.get(tag, self) is not self
  __contains__ = contains

  def __iter__(self):
    elem, prefix = self.__elem, self.__prefix
    pl = len(prefix)
    return iter(t[pl:] for t in elem.getTaggedValueTags() if t.startswith(prefix))

  def list(self): return list(iter(self))

  def unwrap(self): return self.__elem


# the simplest tagger (without prefix)
tag = Tagger()
